/*
** File: debug.c  
** Project: VxWorks Driver Framework
** Purpose: Debugging functions
**
** (C) Copyright Alpha Data 2010
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#if (_WRS_VXWORKS_MAJOR > 6) || (_WRS_VXWORKS_MAJOR == 6 && _WRS_VXWORKS_MINOR >= 6)
# include <vxCpuLib.h>
#endif

#include "df.h"
#include "dfvxworks.h"

#if DF_DBG_BUILD

# define DFVXW_LOG_DELAY_TICKS (0)

# define DFVXW_LOG_MAXNUMMSG  (256)
# define DFVXW_LOG_MAXMSGSIZE (256)
# define DFVXW_LOG_BUFSIZE    (DFVXW_LOG_MAXMSGSIZE * DFVXW_LOG_MAXNUMMSG)

int g_dfDebugLevel = 0;

static char* g_pLogMsgBuf = NULL;
static unsigned int g_logMsgPos = 0;
static DfSpinLock g_logLock;

void
dfVxwLogInit(void)
{
  dfSpinLockInit(&g_logLock);
  g_pLogMsgBuf = (char*)dfMalloc(DFVXW_LOG_BUFSIZE);
  if (NULL == g_pLogMsgBuf) {
#if DFVXW_HAS_VX_USR_ARG_T
    logMsg("*** %s: failed to allocate log message buffer. No further debug messages will be displayed.\n",
      (_Vx_usr_arg_t)g_dfDriverName, (_Vx_usr_arg_t)0, (_Vx_usr_arg_t)0, (_Vx_usr_arg_t)0, (_Vx_usr_arg_t)0, (_Vx_usr_arg_t)0);
#else
    logMsg("*** %s: failed to allocate log message buffer. No further debug messages will be displayed.\n",
      (int)g_dfDriverName, 0, 0, 0, 0, 0);
#endif
  }
}

void
dfVxwLogUninit(void)
{
  if (NULL != g_pLogMsgBuf) {
    dfFree(g_pLogMsgBuf);
    g_pLogMsgBuf = NULL;
  }
}

void
_dfDebugPrint(
  const char* fmt, ...)
{
  DfSpinLockFlags f;
  va_list	ap;
  char* pBuffer;
  char* p;
  size_t n, m = DFVXW_LOG_MAXMSGSIZE;
  unsigned int pos;

  if (NULL == g_pLogMsgBuf) {
    return;
  }

  f = dfSpinLockGet(&g_logLock);
  pos = g_logMsgPos;
  g_logMsgPos = pos + DFVXW_LOG_MAXMSGSIZE;
  if (g_logMsgPos >= DFVXW_LOG_BUFSIZE) {
    g_logMsgPos = 0;
  }
  dfSpinLockPut(&g_logLock, f);
  pBuffer = g_pLogMsgBuf + pos;

  p = pBuffer;
  n = strlen(g_dfDriverName);
  if (n >= m) {
    n = m;
  }
  memcpy(p, g_dfDriverName, n);
  p += n;
  m -= n;
#if (_WRS_VXWORKS_MAJOR > 6) || (_WRS_VXWORKS_MAJOR == 6 && _WRS_VXWORKS_MINOR >= 6)
  /*
  ** In VxWorks 6.6 and later we can get the index of the CPU on which we are
  **  executing. This can't be considered 100% accurate because during execution
  ** of this function we could conceivably be preempted by another task, and
  ** migrated to another CPU.
  */
  if (m) {
    int cpuIndex = vxCpuIndexGet();
    n = snprintf(p, m, "(%d)", cpuIndex);
    p += n;
    m -= n;
  }
#endif
  if (m) {
    *p++ = ':';
    m--;
  }
  if (m) {
    *p++ = ' ';
    m--;
  }
  if (m) {
    va_start(ap, fmt);
#if (_WRS_VXWORKS_MAJOR >= 6)
    /* Need VxWorks 6.x for vsnprintf() */
    vsnprintf(p, m, fmt, ap);
#else
    vsprintf(p, fmt, ap);
#endif
    va_end(ap);
  }
  pBuffer[DFVXW_LOG_MAXMSGSIZE - 1] = '\0';

#if DFVXW_HAS_VX_USR_ARG_T
  logMsg("%s", (_Vx_usr_arg_t)pBuffer, (_Vx_usr_arg_t)0, (_Vx_usr_arg_t)0, (_Vx_usr_arg_t)0, (_Vx_usr_arg_t)0, (_Vx_usr_arg_t)0);
#else
  logMsg("%s", (int)pBuffer, 0, 0, 0, 0, 0);
#endif

#if defined(DFVXW_LOG_DELAY_TICKS) && (DFVXW_LOG_DELAY_TICKS > 0)
  if (!intContext()) {
    taskDelay(DFVXW_LOG_DELAY_TICKS);
  }
#endif
}

DF_DECLARE_INLINE_FUNC(size_t, remaining)(
  const char* pPos,
  const char* line,
  size_t max)
{
  size_t n = pPos - line;
  if (n > max) {
    return 0;
  } else {
    return max - n;
  }
}

void
dfDebugDumpMemory(
  int level,
  unsigned int indent,
  unsigned int nColumn,
  unsigned int radix,
  boolean_t bAlign,
  const void* pBuffer,
  size_t n)
{
  uint64_t address, limit, addressCol, m;
  unsigned int addressLo, columnMask;
  uint8_t* p;
  uint8_t* p8;
  char line[257];
  uint8_t columnBuf[256];
  char* pPos;
  uint8_t val8;
  uint16_t val16;
  uint32_t val32;
  uint64_t val64;
  unsigned int i, j, k, chunk, rowBytes;

  if (sizeof(void*) != 4 && sizeof(void*) != 8) {
    dfDebugPrint(0, ("*** dfDebugDumpMemory: strange value '%lu' for 'sizeof(void*)'\n", (unsigned long)sizeof(void*)));
    return;
  }
  if (indent > 128) {
    dfDebugPrint(0, ("*** dfDebugDumpMemory: invalid value '%lu' for indent\n", (unsigned long)indent));
    return;
  }
  switch (nColumn) {
  case 1:
  case 2:
  case 4:
  case 8:
  case 16:
  case 32:
    break;

  default:
    dfDebugPrint(0, ("*** dfDebugDumpMemory: invalid value '%lu' for nColumn\n", (unsigned long)nColumn));
    return;
  }
  if (radix != 1 && radix != 2 && radix != 4 && radix != 8) {
    dfDebugPrint(0, ("*** dfDebugDumpMemory: invalid value '%lu' for radix\n", (unsigned long)radix));
    return;
  }

  rowBytes = nColumn * radix;

  /* Make sure everything is aligned */
  columnMask = rowBytes - 1U;
  address = (uint64_t)(uintptr_t)pBuffer;
  limit = address + n;
  if (bAlign) {
    addressCol = address & ~(uint64_t)columnMask;
    addressLo = ((unsigned int)address) & columnMask;
  } else {
    addressCol = address;
    addressLo = 0;
  }

  m = limit - address;
  p = (uint8_t*)pBuffer;

  pPos = line;
  memset(pPos, ' ', indent);
  pPos += indent;
#if (_WRS_VXWORKS_MAJOR >= 6)
  snprintf(pPos, remaining(pPos, line, sizeof(line)), "Dump of memory at %p + %lu(0x%lx) bytes:\n",
    pBuffer, (unsigned long)n, (unsigned long)n);
#else
  /* Need VxWorks 6.x for snprintf() */
  sprintf(pPos, "Dump of memory at %p + %lu(0x%lx) bytes:\n",
    pBuffer, (unsigned long)n, (unsigned long)n);
#endif
  dfDebugPrint(level, ("%s", line));

  pPos = line;
  if (sizeof(void*) == 4) {
    memset(pPos, ' ', indent + 12);
    pPos += indent + 12;
  } else {
    memset(pPos, ' ', indent + 21);
    pPos += indent + 21;
  }
  for (i = 0; i < nColumn; i += radix) {
    switch (radix) {
    case 0x1U:
#if (_WRS_VXWORKS_MAJOR >= 6)
      snprintf(pPos, remaining(pPos, line, sizeof(line)), "%02lx ", (unsigned long)i);
#else
      /* Need VxWorks 6.x for snprintf() */
      sprintf(pPos, "%02lx ", (unsigned long)i);
#endif
      pPos += 3;
      break;

    case 0x2U:
#if (_WRS_VXWORKS_MAJOR >= 6)
      snprintf(pPos, remaining(pPos, line, sizeof(line)), "  %02lx ", (unsigned long)i);
#else
      /* Need VxWorks 6.x for snprintf() */
      sprintf(pPos, "  %02lx ", (unsigned long)i);
#endif
      pPos += 5;
      break;

    case 0x4U:
#if (_WRS_VXWORKS_MAJOR >= 6)
      snprintf(pPos, remaining(pPos, line, sizeof(line)), "      %02lx ", (unsigned long)i);
#else
      /* Need VxWorks 6.x for snprintf() */
      sprintf(pPos, "      %02lx ", (unsigned long)i);
#endif
      pPos += 9;
      break;

    case 0x8U:
#if (_WRS_VXWORKS_MAJOR >= 6)
      snprintf(pPos, remaining(pPos, line, sizeof(line)), "               %02lx ", (unsigned long)i);
#else
      /* Need VxWorks 6.x for snprintf() */
      sprintf(pPos, "               %02lx ", (unsigned long)i);
#endif
      pPos += 18;
      break;
    }
  }
  *pPos++ = '\n';
  *pPos++ = '\0';
  dfDebugPrint(level, ("%s", line));

  while (m) {
    chunk = rowBytes - addressLo;
    if (chunk > m) {
      chunk = (unsigned int)m;
    }

    memset(columnBuf, 0, rowBytes);
    memcpy(columnBuf + addressLo, p, chunk);

    p8 = (uint8_t*)columnBuf;

    pPos = line;
    memset(pPos, ' ', indent);
    pPos += indent;
    if (sizeof(void*) == 4) {
#if (_WRS_VXWORKS_MAJOR >= 6)
      snprintf(pPos, remaining(pPos, line, sizeof(line)), "0x%08lx: ", (unsigned long)(uint32_t)addressCol);
#else
      /* Need VxWorks 6.x for snprintf() */
      sprintf(pPos, "0x%08lx: ", (unsigned long)(uint32_t)addressCol);
#endif
      pPos += 12;
    } else {
#if (_WRS_VXWORKS_MAJOR >= 6)
      snprintf(pPos, remaining(pPos, line, sizeof(line)), "0x%08lx_%08lx: ", (unsigned long)(uint32_t)(addressCol >> 32U), (unsigned long)(uint32_t)addressCol);
#else
      /* Need VxWorks 6.x for snprintf() */
      sprintf(pPos, "0x%08lx_%08lx: ", (unsigned long)(uint32_t)(addressCol >> 32U), (unsigned long)(uint32_t)addressCol);
#endif
      pPos += 21;
    }

    j = addressLo + chunk;
    for (i = 0; i < rowBytes; i += radix) {
      switch (radix) {
      case 0x1U:
        val8 = *p8++;
        if (i >= addressLo && i < j) {
#if (_WRS_VXWORKS_MAJOR >= 6)
          snprintf(pPos, remaining(pPos, line, sizeof(line)), "%02lx", (unsigned long)val8);
#else
          /* Need VxWorks 6.x for snprintf() */
          sprintf(pPos, "%02lx", (unsigned long)val8);
#endif
          pPos += 2;
        } else {
          *pPos++ = '.';
          *pPos++ = '.';
        }
        break;

      case 0x2U:
        val16 = dfReadUnaligned16((uint16_t*)p8);
        val16 = dfLe32ToCpu(val16);
        p8 += 2;
        for (k = 2; k != 0;) {
          k--;
          if (i + k >= addressLo && i + k < j) {
#if (_WRS_VXWORKS_MAJOR >= 6)
            snprintf(pPos, remaining(pPos, line, sizeof(line)), "%02lx", (unsigned long)(uint8_t)(val16 >> (k << 3)));
#else
            /* Need VxWorks 6.x for snprintf() */
            sprintf(pPos, "%02lx", (unsigned long)(uint8_t)(val16 >> (k << 3)));
#endif
            pPos += 2;
          } else {
            *pPos++ = '.';
            *pPos++ = '.';
          }
          if (k == 3) {
            *pPos++ = '_';
          }
        }
        break;

      case 0x4U:
        val32 = dfReadUnaligned32((uint32_t*)p8);
        val32 = dfLe32ToCpu(val32);
        p8 += 4;
        for (k = 4; k != 0;) {
          k--;
          if (i + k >= addressLo && i + k < j) {
#if (_WRS_VXWORKS_MAJOR >= 6)
            snprintf(pPos, remaining(pPos, line, sizeof(line)), "%02lx", (unsigned long)(uint8_t)(val32 >> (k << 3)));
#else
            /* Need VxWorks 6.x for snprintf() */
            sprintf(pPos, "%02lx", (unsigned long)(uint8_t)(val32 >> (k << 3)));
#endif
            pPos += 2;
          } else {
            *pPos++ = '.';
            *pPos++ = '.';
          }
        }
        break;

      case 0x8U:
        val64 = dfReadUnaligned64((uint64_t*)p8);
        val64 = dfLe64ToCpu(val64);
        p8 += 8;
        for (k = 8; k != 0;) {
          k--;
          if (i + k >= addressLo && i + k < j) {
#if (_WRS_VXWORKS_MAJOR >= 6)
            snprintf(pPos, remaining(pPos, line, sizeof(line)), "%02lx", (unsigned long)(uint8_t)(val64 >> (k << 3)));
#else
            /* Need VxWorks 6.x for snprintf() */
            sprintf(pPos, "%02lx", (unsigned long)(uint8_t)(val64 >> (k << 3)));
#endif
            pPos += 2;
          } else {
            *pPos++ = '.';
            *pPos++ = '.';
          }
          if (k == 4) {
            *pPos++ = '_';
          }
        }
        break;
      }
      *pPos++ = ' ';
    }

    p8 = (uint8_t*)columnBuf;
    for (i = 0; i < rowBytes; i++) {
      char c = (char)*p8++;

      if (isprint((int)c)) {
        *pPos++ = c;
      } else {
        *pPos++ = '.';
      }
    }
    *pPos++ = '\n';
    *pPos++ = '\0';
    dfDebugPrint(level, (line));
    addressLo = 0;
    addressCol += rowBytes;
    m -= chunk;
    p += chunk;
  }
}

# if defined(DF_TRACK_MALLOC)

void*
_dfVxwDbgMalloc(
  size_t n,
  const char* pFilename,
  int lineNum)
{
  void* p;

  p = malloc(n);
  if (NULL == p) {
    return NULL;
  }
  bzero(p, n);

#if DFVXW_HAS_VX_USR_ARG_T
  logMsg("%s: (malloc) %s:%d n=0x%d(0x%x) ptr=0x%x\n", (_Vx_usr_arg_t)g_dfDriverName, (_Vx_usr_arg_t)pFilename, lineNum, (_Vx_usr_arg_t)n, (_Vx_usr_arg_t)n, (_Vx_usr_arg_t)p);
#else
  logMsg("%s: (malloc) %s:%d n=0x%d(0x%x) ptr=0x%x\n", (int)g_dfDriverName, (int)pFilename, lineNum, (int)n, (int)n, (int)p);
#endif
  return p;
}

void
_dfVxwDbgFree(
  void* p,
  const char* pFilename,
  int lineNum)
{
#if DFVXW_HAS_VX_USR_ARG_T
  logMsg("%s: (free) %s:%d ptr=0x%x\n", (_Vx_usr_arg_t)g_dfDriverName, (_Vx_usr_arg_t)pFilename, lineNum, (_Vx_usr_arg_t)p, 0, 0);
#else
  logMsg("%s: (free) %s:%d ptr=0x%x\n", (int)g_dfDriverName, (int)pFilename, lineNum, (int)p, 0, 0);
#endif
  free(p);
}

# endif

#endif

extern void
dfVxwCheckDataTypes(
  void)
{
  dfAssert(sizeof(uint8_t) == 1);
  dfAssert(sizeof(uint16_t) == 2 * sizeof(uint8_t));
  dfAssert(sizeof(uint32_t) == 4 * sizeof(uint8_t));
  dfAssert(sizeof(uint32_t) >= sizeof(unsigned int));
  dfAssert(sizeof(uint64_t) == 8 * sizeof(uint8_t));
  dfAssert(sizeof(uint64_t) <= sizeof(unsigned long long));
  dfAssert(sizeof(uint64_t) >= sizeof(size_t));
  dfAssert(sizeof(int8_t) == 1);
  dfAssert(sizeof(int16_t) == 2 * sizeof(int8_t));
  dfAssert(sizeof(int32_t) == 4 * sizeof(int8_t));
  dfAssert(sizeof(int32_t) >= sizeof(int));
  dfAssert(sizeof(int64_t) == 8 * sizeof(uint8_t));
  dfAssert(sizeof(size_t) == sizeof(void*));
  dfAssert(sizeof(size_t) >= sizeof(unsigned long));
  dfAssert(sizeof(uintptr_t) == sizeof(void*));
#if (_WRS_VXWORKS_MAJOR > 6) || (_WRS_VXWORKS_MAJOR == 6 && _WRS_VXWORKS_MINOR >= 6)
  dfAssert(sizeof(atomicVal_t) >= sizeof(int));
#endif
}
