/*
** File: mapio.c  
** Project: VxWorks Driver Framework
** Purpose: Mechanism by which OS-independent driver code can map device memory
**          into the driver's address space.
**
** (C) Copyright Alpha Data 2010
*/

#include "df.h"
#include "dfvxworks.h"

#if defined(ADB3_VXBUS)

# if defined(DFVXW_VXBUS_GEN2)

/* VxBus Gen 2 version */
void*
dfIoSpaceMap(
  DfDeviceObject* pDevObj,
  DfBusResources* pResources,
  unsigned int barIndex,
  uint64_t offset,
  uint64_t length,
  DfMemoryHandle* pHandle)
{
  VXB_DEV_ID pVxbDev = pDevObj->system.pVxbDev;
  VXB_RESOURCE_LIST* pVxbResourceList = vxbResourceListGet(pVxbDev);
  VXB_RESOURCE* pVxbResource;
  VXB_RESOURCE* pTmp;
  VXB_RESOURCE_ADR* pVxbResourceAdr;
  unsigned int i, n;
  void* pKernelAddress;

  if (NULL == pVxbResourceList) {
    return NULL; /* No resources!? */
  }

  for (i = 0, n = 0; i < VXB_MAXBARS; i++) {
    pVxbResource = vxbResourceFind(pVxbResourceList, VXB_RES_ID_CREATE(VXB_RES_MEMORY, i));
    if (NULL == pVxbResource) {
      continue;
    }

    if (n == barIndex) {
      pTmp = vxbResourceAlloc(pVxbDev, VXB_RES_MEMORY, (UINT16)i);
      if (pTmp == pVxbResource) {
        pVxbResourceAdr = (VXB_RESOURCE_ADR*)pVxbResource->pRes;
        *pHandle = pVxbResourceAdr->pHandle;
        pKernelAddress = (void*)(pVxbResourceAdr->virtAddr + offset);
        return pKernelAddress;
      }
    }

    n++;
  }

  return NULL; /* Not found */
}

# else

/* VxBus 3/4/5 version */
void*
dfIoSpaceMap(
  DfDeviceObject* pDevObj,
  DfBusResources* pResources,
  unsigned int barIndex,
  uint64_t offset,
  uint64_t length,
  DfMemoryHandle* pHandle)
{
  VXB_DEVICE_ID pVxbDev = pDevObj->system.pVxbDev;
  DfMemoryHandle handle;
  unsigned int n = barIndex;
  int i;

  /* Find the requested memory BAR's index in the VxBus device structure; may not be the same as 'barIndex' */
  for (i = 0; i < VXB_MAXBARS; i++) {
    if (pVxbDev->regBaseFlags[i] == VXB_REG_MEM) {
      if (0 == n) {
        break;
      }
      n--;
    }
  }
  if (OK == vxbRegMap(pVxbDev, (int)i, &handle)) {
    *pHandle = handle;
    return pVxbDev->pRegBase[i];
  } else {
    return NULL;
  }
}

# endif

#else

/* Legacy version */
void*
dfIoSpaceMap(
  DfDeviceObject* pDevObj,
  DfBusResources* pResources,
  unsigned int barIndex,
  uint64_t offset,
  uint64_t length,
  DfMemoryHandle* pHandle)
{
  void* pKernelAddress = NULL;
  DfPciBar* pBar;

  /* Sanity checks */
  dfAssert(length < (unsigned long)-1);
  dfAssert(offset + length >= offset);
  dfAssert(pDevObj->hardwareId.busType == DfBusPci);
  dfAssert(barIndex < pResources->variant.pci.numMemBar);
  pBar = &pResources->variant.pci.memBar[barIndex];
  dfAssert(offset < pBar->size);
  dfAssert(offset + length <= pBar->size);

  if (!dfVxwLegacyBusToCpuAddress(pDevObj, pBar->base + offset, &pKernelAddress)) {
    return NULL;
  }

  *pHandle = NULL; /* handle is always NULL in legacy driver */
  return pKernelAddress;
}

#endif

#if defined(DFVXW_VXBUS_GEN2)

/* VxBus Gen 2 version */
boolean_t
dfIoSpaceUnmap(
  DfDeviceObject* pDevObj,
  void* pKernelAddress)
{
  VXB_DEV_ID pVxbDev = pDevObj->system.pVxbDev;
  VXB_RESOURCE_LIST* pVxbResourceList = vxbResourceListGet(pVxbDev);
  VXB_RESOURCE* pVxbResource;
  VXB_RESOURCE_ADR* pVxbResourceAdr;
  VIRT_ADDR virtBase, virtLimit;
  int i;

  if (NULL == pVxbResourceList) {
    return FALSE; /* No resources!? */
  }

  for (i = 0; i < VXB_MAXBARS; i++) {
    pVxbResource = vxbResourceFind(pVxbResourceList, VXB_RES_ID_CREATE(VXB_RES_MEMORY, i));
    if (NULL == pVxbResource) {
      continue;
    }

    pVxbResourceAdr = (VXB_RESOURCE_ADR*)pVxbResource->pRes;
    virtBase = pVxbResourceAdr->virtAddr;
    virtLimit = pVxbResourceAdr->virtAddr + pVxbResourceAdr->size;
    if (virtBase <= (VIRT_ADDR)pKernelAddress && (VIRT_ADDR)pKernelAddress < virtLimit) {
      (void)vxbResourceFree(pVxbDev, pVxbResource);
      return TRUE; /* Found and unmapped */
    }    
  }

  return FALSE; /* Not found */
}


#else

/* VxBus 3/4/5 version */
/* Legacy version */
boolean_t
dfIoSpaceUnmap(
  DfDeviceObject* pDevObj,
  void* pKernelAddress)
{
  dfAssert(NULL != pKernelAddress);

  /* Nothing to do */

  return TRUE;
}

#endif
