/*
** File: model_adpexrc5t.c
** Project: ADMXRC3 module driver
** Purpose: ADPe-XRC-5T specific code, including the function to bootstrap
**          the device context.
**
** (C) Copyright Alpha Data 2009
*/

#include <df.h>

#include <core/adpexrc5t.h>
#include "device.h"
#include "info.h"
#include "model_boot.h"

static ADMXRC3_STATUS
getBankInfo(
	Admxrc3DeviceContext* pDevCtx,
  unsigned int bankIndex,
  const void* pVpdBuffer,
  IOCTLS_ADMXRC3_GETBANKINFO* pIoctl)
{
  VpdAdpexrc5tRev0* pVpd = (VpdAdpexrc5tRev0*)pVpdBuffer;
  uint8_t order;

  if (bankIndex >= 6) {
    return ADMXRC3_INVALID_INDEX;
  }
  if (bankIndex < 4) {
    order = pVpd->sdram[bankIndex].order;
    pIoctl->out.typeMask = ADMXRC3_BANK_SDRAM_DDR2;
    pIoctl->out.physicalDataWidth = 32;
    pIoctl->out.physicalEccWidth = 0;
    pIoctl->out.maximumFrequency = (uint64_t)pVpd->sdram[bankIndex].speed * 100000;
    pIoctl->out.minimumFrequency = 125000000U;
  } else {
    order = pVpd->ssram[bankIndex - 4].order;
    pIoctl->out.typeMask = ADMXRC3_BANK_SSRAM_QDR;
    pIoctl->out.physicalDataWidth = 16;
    pIoctl->out.physicalEccWidth = 2;
    pIoctl->out.maximumFrequency = (uint64_t)pVpd->ssram[bankIndex - 4].speedMax * 100000;
    pIoctl->out.minimumFrequency = (uint64_t)pVpd->ssram[bankIndex - 4].speedMin * 100000;
  }
  pIoctl->out.connectivityMask = 0x1U;
  pIoctl->out.physicalSize = (uint64_t)0x1U << order;
  pIoctl->out.bPresent = (uint8_t)(order != 0);

  return ADMXRC3_SUCCESS;
}

static void
getCardInfo(
	Admxrc3DeviceContext* pDevCtx,
  const void* pVpdBuffer,
  IOCTLS_ADMXRC3_GETCARDINFO* pIoctl)
{
  VpdAdpexrc5tRev0* pVpd = (VpdAdpexrc5tRev0*)pVpdBuffer;
  unsigned int i;

  pIoctl->out.serialNumber = pVpd->serial;
  pIoctl->out.numMemoryBank = 6;
  pIoctl->out.memoryBankPresent = 0;
  for (i = 0; i < 4; i++) {
    if (0 != pVpd->sdram[i].order) {
      /* Bank is fitted */
      pIoctl->out.memoryBankPresent |= (0x1U << i);
    }
  }
  for (; i < 6; i++) {
    if (0 != pVpd->ssram[i - 4].order) {
      /* Bank is fitted */
      pIoctl->out.memoryBankPresent |= (0x1U << i);
    }
  }
}

static ADMXRC3_STATUS
getFpgaInfo(
	Admxrc3DeviceContext* pDevCtx,
  unsigned int targetIndex,
  const void* pVpdBuffer,
  IOCTLS_ADMXRC3_GETFPGAINFO* pIoctl)
{
  VpdAdpexrc5tRev0* pVpd = (VpdAdpexrc5tRev0*)pVpdBuffer;
  ADMXRC3_PACKAGE_TYPE package;
  ADMXRC3_FPGA_TYPE fpgaType;

  switch (targetIndex) {
  case 0:
    package = ADMXRC3_PACKAGE_MAKEFF(1738);
    fpgaType = (ADMXRC3_FPGA_TYPE)pVpd->fpga.device;
    break;

  default:
    return ADMXRC3_INVALID_INDEX;
  }

  pIoctl->out.flags = 0;
  setupIoctlsGetFpgaInfo(pDevCtx, pIoctl, fpgaType, package, NULL, NULL, TRUE);

  return ADMXRC3_SUCCESS;
}

void
admxrc3BootstrapAdpexrc5t(
	Admxrc3DeviceContext* pDevCtx)
{
  dfDebugPrint(1, ("admxrc3BootstrapAdpexrc5t: entered\n"));

  pDevCtx->methods.pGetBankInfo = getBankInfo;
  pDevCtx->methods.pGetCardInfo = getCardInfo;
  pDevCtx->methods.pGetFpgaInfo = getFpgaInfo;
  pDevCtx->bReverseSelectMapData = FALSE;
  pDevCtx->localAddressLimit = (uint64_t)(0x1ULL << 39U);
  pDevCtx->localWidth = 16U;
}
